/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects
 * @author     phesch ( ulila at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "predictions.h"

#include <SimoxUtility/math/pose/pose.h>
#include <SimoxUtility/math/regression/linear.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/ice_conversions/ice_conversions_templates.h>

namespace armarx::objpose
{

    objpose::ObjectPosePredictionResult
    predictObjectPoseLinear(const std::map<DateTime, ObjectPose>& poses,
                      const DateTime& time,
                      const ObjectPose& latestPose)
    {
        objpose::ObjectPosePredictionResult result;
        result.success = false;

        const DateTime timeOrigin = DateTime::Now();

        std::vector<double> timestampsSec;
        std::vector<Eigen::Vector3d> positions;

        // ToDo: How to handle attached objects?
        for (const auto& [timestamp, pose] : poses)
        {
            timestampsSec.push_back((timestamp - timeOrigin).toSecondsDouble());
            positions.emplace_back(simox::math::position(pose.objectPoseGlobal).cast<double>());
        }

        ARMARX_CHECK_EQUAL(timestampsSec.size(), positions.size());

        Eigen::Vector3d prediction;
        // Static objects don't move. Objects that haven't moved for a while probably won't either.
        if (timestampsSec.size() <= 1 || latestPose.isStatic)
        {
            prediction = simox::math::position(latestPose.objectPoseGlobal).cast<double>();
        }
        else
        {
            using simox::math::LinearRegression3d;
            const bool inputOffset = false;

            const LinearRegression3d model =
                LinearRegression3d::Fit(timestampsSec, positions, inputOffset);
            const auto predictionTime = armarx::fromIce<DateTime>(time);
            prediction = model.predict((predictionTime - timeOrigin).toSecondsDouble());
        }

        // Used as a template for the returned pose prediction.
        ObjectPose latestCopy = latestPose;

        // Reuse the rotation from the latest pose.
        // This is a pretty generous assumption, but the linear model doesn't cover rotations,
        // so it's our best guess.
        Eigen::Matrix4f latest = latestPose.objectPoseGlobal;
        simox::math::position(latest) = prediction.cast<float>();
        latestCopy.setObjectPoseGlobal(latest);

        result.success = true;
        result.prediction = latestCopy.toIce();

        return result;
    }
} // namespace armarx::objpose
