#pragma once

#include <RobotAPI/libraries/aron/core/data/variant/All.h>
#include <RobotAPI/libraries/aron/core/type/variant/All.h>


namespace armarx::aron::visitor
{

    /**
     * @brief A visitor traversing an `aron::type::dto::GenericTypePtr` using
     * `aron::Typenavigator::Navigator`s, taking care of type checking,
     * casting and hierarchical traversal.
     *
     * To use it, create a class deriving from `AronTypeVisitor` and
     * override the overloads of `visitEnter()`, `visitExit()` and `visit()`
     * for the types you would like to handle (see blow for a copy-and-paste
     * example).
     * Then, instantiate your derived visitor and call `applyTo()`, passing
     * the `AronTypePtr` or the root `Navigator`.
     *
     * For dicts and lists, you will get a `visitEnter()`/`visitExit()` pair.
     * For types without children (e.g. primitive types), you get a single
     * `visit()` call.
     *
     * Each visit function should return true to continue traversal or
     * false to stop it. All default implementations return true.
     *
     * To the the location of the passed navigator in the original `AronType`,
     * use `navigator.getPath()`.
     *
     * @see `aron::type::dto::GenericTypePtr`
     * @see `aron::Typenavigator::Navigator`
     * @see `std::stack` To manage Type with stack semantics.
     */
    class TypedDataVisitor
    {
    public:

        // Define types to ease writing code in this and derived classes.

        // Data - possible data representations
        using DataNavigator = data::Variant;

        using DictDataNavigator = data::Dict;
        using ListDataNavigator = data::List;

        using BoolDataNavigator = data::Bool;
        using DoubleDataNavigator = data::Double;
        using FloatDataNavigator = data::Float;
        using IntDataNavigator = data::Int;
        using LongDataNavigator = data::Long;
        using StringDataNavigator = data::String;

        using NDArrayDataNavigator = data::NDArray;


        // Type
        using TypeNavigator = type::Variant;

        // Dict-valued
        using DictTypeNavigator = type::Dict;
        using ObjectTypeNavigator = type::Object;

        // List-valued
        using ListTypeNavigator = type::List;
        using TupleTypeNavigator = type::Tuple;

        // Primitive-valued
        using BoolTypeNavigator = type::Bool;
        using DoubleTypeNavigator = type::Double;
        using FloatTypeNavigator = type::Float;
        using IntTypeNavigator = type::Int;
        using LongTypeNavigator = type::Long;
        using StringTypeNavigator = type::String;
        using TimeTypeNavigator = type::Time;


        // Array-valued
        using EigenMatrixTypeNavigator = type::Matrix;
        using EigenQuaternionTypeNavigator = type::Quaternion;
        using ImageNavigator = type::Image;
        using PCLPointCloudTypeNavigator = type::PointCloud;
        using PoseTypeNavigator = type::Pose;
        using PositionTypeNavigator = type::Position;
        using OrientationTypeNavigator = type::Orientation;


    public:

        TypedDataVisitor();
        virtual ~TypedDataVisitor();


        bool applyTo(TypeNavigator& type, DataNavigator& data);
        bool applyTo(const std::string& key, TypeNavigator& type, DataNavigator& data);


        virtual bool visitEnter(DictTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(DictTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(ObjectTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(ObjectTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(ListTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(ListTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(TupleTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(TupleTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }


        virtual bool visit(BoolTypeNavigator& type, BoolDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(DoubleTypeNavigator& type, DoubleDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(FloatTypeNavigator& type, FloatDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(IntTypeNavigator& type, IntDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(LongTypeNavigator& type, LongDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(StringTypeNavigator& type, StringDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(TimeTypeNavigator& type, LongDataNavigator& data)
        {
            return visit(type, IceUtil::Time::microSeconds(data.getValue()));
        }
        virtual bool visit(TimeTypeNavigator& type, const IceUtil::Time& data)
        {
            (void) type, (void) data;
            return true;
        }


        virtual bool visit(EigenMatrixTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(EigenQuaternionTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(ImageNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        /*virtual bool visit(IVTCByteImageTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(OpenCVMatTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }*/
        virtual bool visit(PCLPointCloudTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(PoseTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(PositionTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(OrientationTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }



        virtual bool visitEnter(DictTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            return visitEnter(type, data);
        }
        virtual bool visitExit(DictTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            return visitExit(type, data);
        }

        virtual bool visitEnter(ObjectTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            return visitEnter(type, data);
        }
        virtual bool visitExit(ObjectTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            return visitExit(type, data);
        }

        virtual bool visitEnter(ListTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            return visitEnter(type, data);
        }
        virtual bool visitExit(ListTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            return visitExit(type, data);
        }

        virtual bool visitEnter(TupleTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            return visitEnter(type, data);
        }
        virtual bool visitExit(TupleTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            return visitExit(type, data);
        }


        virtual bool visit(BoolTypeNavigator& type, const std::string& key, BoolDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(DoubleTypeNavigator& type, const std::string& key, DoubleDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(FloatTypeNavigator& type, const std::string& key, FloatDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(IntTypeNavigator& type, const std::string& key, IntDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(LongTypeNavigator& type, const std::string& key, LongDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(StringTypeNavigator& type, const std::string& key, StringDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(TimeTypeNavigator& type, const std::string& key, LongDataNavigator& data)
        {
            (void) key;
            return visit(type, data);
        }
        virtual bool visit(TimeTypeNavigator& type, const std::string& key, const IceUtil::Time& data)
        {
            (void) key;
            return visit(type, data);
        }


        virtual bool visit(EigenMatrixTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(EigenQuaternionTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(ImageNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        /*virtual bool visit(IVTCByteImageTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(OpenCVMatTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }*/
        virtual bool visit(PCLPointCloudTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(PoseTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(PositionTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }
        virtual bool visit(OrientationTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            return visit(type, data);
        }


    private:

        template <class DataT, class TypeT>
        bool applyToCollection(TypeT& type, const std::string& key, DataNavigator& data);

        bool applyToChildren(ObjectTypeNavigator& type, DictDataNavigator& data);
        bool applyToChildren(DictTypeNavigator& type, DictDataNavigator& data);
        bool applyToChildren(ListTypeNavigator& type, ListDataNavigator& data);
        bool applyToChildren(TupleTypeNavigator& type, ListDataNavigator& data);

    };

}
