#pragma once

#include <mutex>
#include <queue>
#include <thread>
#include <functional>

#include <ArmarXCore/core/ComponentPlugin.h>
#include <ArmarXCore/core/ManagedIceObject.h>

#include <ArmarXCore/core/services/tasks/RunningTask.h>

#include <RobotAPI/interface/skills/SkillManagerInterface.h>
#include <RobotAPI/libraries/aron/core/data/variant/All.h>

#include "SkillDescription.h"

namespace armarx
{
    namespace skills
    {
        class Skill : public armarx::Logging
        {
        public:
            using CallbackT = std::function<void(const aron::data::DictPtr&)>;

            enum class Status
            {
                Succeeded,
                TimeoutReached,
                Stopped,
                Failed
            };

            Skill() = delete;
            Skill(const SkillDescription&);
            virtual ~Skill() = default;

            /// Override this method with the actual implementation.
            virtual void init(const aron::data::DictPtr& params);

            /// Override this method with the actual implementation.
            virtual void exit(const aron::data::DictPtr& params);

            /// Override this method with the actual implementation. The callback is for status updates to the calling instance
            virtual Status execute(const aron::data::DictPtr& params, const CallbackT& callback = [](const aron::data::DictPtr& returnValue) { (void) returnValue; });

            /// Do all methods at once.
            Status initExecuteExit(const aron::data::DictPtr& params, const CallbackT& callback = [](const aron::data::DictPtr& returnValue) { (void) returnValue; });

            /// Override if you have special members that needs to be resetted. It is called before the skill ititializes
            virtual void reset();

            /// Override these methods if you want to do something special when notification comes
            virtual void notifyTimeoutReached();
            virtual void notifyStopped();

            virtual bool stopRequested() const;

        private:
            /// install a condition as a seperate thread
            std::thread installTimeoutCondition();

        public:
            const SkillDescription description;

            /// running params
            IceUtil::Time started = IceUtil::Time::milliSeconds(-1);
            IceUtil::Time exited = IceUtil::Time::milliSeconds(-1);

            /// proxies that called the skills. Will be set from provider and is const afterwards
            provider::dti::SkillProviderInterfacePrx ownerProvider = nullptr;
            manager::dti::SkillManagerInterfacePrx ownerManager = nullptr;

        protected:
            /// Use conditions during runtime this way
            std::atomic_bool running = true;
            std::atomic_bool stopped = false;
            std::atomic_bool timeoutReached = false;

        private:
            std::thread timeoutCheck;
        };
    }
}
